@@ -3,50 +3,31 @@
 # SPDX-License-Identifier: MIT
 
 """
-Information and methods pertaining to links and link files
+Information and methods pertaining to links and link files.
 
 Author(s): Alec Delaney (Tekktrik)
 """
 
-import os
-import time
 import json
+import os
 import pathlib
 import shutil
-import csv
-import functools
-import fcntl
-from collections import namedtuple
-from typing import Dict, List, Union, Iterator, Optional, Literal
-from typer import get_app_dir, Exit
+import time
+from typing import Dict, List, Tuple, TypeAlias, Union
 
+from typer import Exit
 
-# Filepath constants
-APP_DIRECTORY = get_app_dir("circlink")
-LINKS_DIRECTORY = os.path.join(APP_DIRECTORY, "links")
-LEDGER_FILE = os.path.join(APP_DIRECTORY, "ledger.csv")
-
-# Namedtuple for ledger entries
-LedgerEntry = namedtuple("LedgerEntry", ("filename", "link_id", "process_id"))
-
-
-def ensure_links_folder() -> None:
-    """Ensure the links folder is created"""
-
-    if not os.path.exists(LINKS_DIRECTORY):
-        os.mkdir(LINKS_DIRECTORY)
-
-
-def ensure_ledger_file() -> None:
-    """Ensure the ledger file exists, or create it if not"""
-
-    ledger_path = pathlib.Path(LEDGER_FILE)
-    ledger_path.touch(exist_ok=True)
+from circlink import LINKS_DIRECTORY
+from circlink.ledger import LedgerEntry, append_to_ledger, remove_from_ledger
 
+# Type aliases
+_TableRowEntry: TypeAlias = Tuple[
+    int, str, bool, pathlib.Path, pathlib.Path, bool, int, str
+]
 
 # pylint: disable=too-many-instance-attributes
 class CircuitPythonLink:
-    """Thelink to the device"""
+    """The link to the device."""
 
     def __init__(
         self,
@@ -63,7 +44,7 @@ class CircuitPythonLink:
         end_flag: bool = False,
         stopped: bool = False,
     ) -> None:
-
+        """Initialize the link."""
         self._read_path = pathlib.Path(read_path)
         self._write_path = pathlib.Path(write_path)
         self._base_dir = pathlib.Path(base_dir)
@@ -86,8 +67,7 @@ class CircuitPythonLink:
 
     @staticmethod
     def get_next_link_id() -> int:
-        """Get the next link ID"""
-
+        """Get the next link ID."""
         link_gen = pathlib.Path(LINKS_DIRECTORY).glob("link*.json")
         link_nums = [int(link_file.name[4:-5]) for link_file in link_gen]
         if not link_nums:
@@ -96,52 +76,51 @@ class CircuitPythonLink:
 
     @property
     def read_path(self) -> pathlib.Path:
-        """The read path for the link"""
+        """Get the read path for the link."""
         return self._read_path
 
     @property
     def write_path(self) -> pathlib.Path:
-        """The write path for the link"""
+        """Get the write path for the link."""
         return self._write_path
 
     @property
     def base_dir(self) -> pathlib.Path:
-        """The base directory for the read path"""
+        """Get the base directory for the read path."""
         return self._base_dir
 
     @property
     def name(self) -> str:
-        """The link name"""
+        """Get the link name."""
         return self._name
 
     @property
     def recursive(self) -> bool:
-        """Whether the link is recursive for the read path"""
+        """Whether the link is recursive for the read path."""
         return self._recursive
 
     @property
     def link_id(self) -> int:
-        """The link ID"""
+        """Link ID."""
         return self._link_id
 
     @property
     def skip_presave(self) -> bool:
-        """Whether a forced save was enacted at the start of the link"""
+        """Whether a forced save was enacted at the start of the link."""
         return self._skip_presave
 
     @property
     def wipe_dest(self) -> bool:
-        """Whether the wrie path was recursively wiped before starting the link"""
+        """Whether the wrie path was recursively wiped before starting the link."""
         return self._wipe_dest
 
     @property
     def stopped(self) -> bool:
-        """Whether the link is marked has stopped"""
+        """Whether the link is marked has stopped."""
         return self._stopped
 
     def save_link(self, *, save_directory: str = LINKS_DIRECTORY) -> pathlib.Path:
-        """Save the link object as a file in the specified folder"""
-
+        """Save the link object as a file in the specified folder."""
         # Create the representative object
         link_obj = {
             "name": self._name,
@@ -171,9 +150,8 @@ class CircuitPythonLink:
 
     @classmethod
     def load_link_by_filepath(cls, link_filepath: str) -> "CircuitPythonLink":
-        """Create a CircuitPythonLink from a JSON file, by filepath"""
-
-        with open(link_filepath, mode="r", encoding="utf-8") as linkfile:
+        """Create a CircuitPythonLink from a JSON file, by filepath."""
+        with open(link_filepath, encoding="utf-8") as linkfile:
             link_obj = json.load(linkfile)
 
         link = cls(
@@ -196,27 +174,25 @@ class CircuitPythonLink:
 
     @classmethod
     def load_link_by_num(cls, link_num: int) -> "CircuitPythonLink":
-        """Create a CircuitPythonLink from a JSON file, by number"""
-
+        """Create a CircuitPythonLink from a JSON file, by number."""
         link_filepath = cls.link_id_to_filename(link_num)
         return cls.load_link_by_filepath(link_filepath)
 
     @staticmethod
     def link_id_to_filename(num: int, *, directory: str = LINKS_DIRECTORY) -> str:
-        """Create a link filename from a link ID"""
+        """Create a link filename from a link ID."""
         return os.path.join(directory, "link" + str(num) + ".json")
 
     @staticmethod
     def filename_to_link_id(filepath: Union[pathlib.Path, str]) -> int:
-        """Get a link ID from a filename"""
-
+        """Get a link ID from a filename."""
         if isinstance(filepath, str):
             filepath = pathlib.Path(filepath)
 
         return int(filepath.name[4:-5])
 
-    def _get_files_monitored(self):
-
+    def get_files_monitored(self):
+        """Get the existing files to be monitored by this link."""
         file_pattern = self._read_path.name
         file_parent = self._read_path.parent
 
@@ -230,8 +206,7 @@ class CircuitPythonLink:
 
     # pylint: disable=too-many-branches
     def begin_monitoring(self) -> None:
-        """Monitor the listed file(s) for changes"""
-
+        """Monitor the listed file(s) for changes."""
         # Ensure the write path exists
         os.makedirs(self._write_path, exist_ok=True)
 
@@ -240,7 +215,7 @@ class CircuitPythonLink:
             shutil.rmtree(self._write_path)
 
         # Get the files that match the read path
-        read_files = self._get_files_monitored()
+        read_files = self.get_files_monitored()
         update_map: Dict[pathlib.Path, float] = {}
 
         # Add all the files to ledger and monitor struct if not already
@@ -268,7 +243,7 @@ class CircuitPythonLink:
             time.sleep(0.1)
 
             # Detect new files
-            read_files = self._get_files_monitored()
+            read_files = self.get_files_monitored()
             new_files: List[pathlib.Path] = []
             for file in read_files:
                 ledger_file_path = str(
@@ -320,7 +295,7 @@ class CircuitPythonLink:
             marked_delete = []
 
         # Remove files from ledger
-        for file in self._get_files_monitored():
+        for file in self.get_files_monitored():
             ledger_entry = LedgerEntry(
                 str(file.resolve()), self.link_id, self.process_id
             )
@@ -335,7 +310,7 @@ class CircuitPythonLink:
     def get_write_filepath(
         write_path: pathlib.Path, read_file: pathlib.Path, base_dir: pathlib.Path
     ) -> pathlib.Path:
-        """Get the write filepath for a specific file"""
+        """Get the write filepath for a specific file."""
         read_file_relative = read_file.relative_to(base_dir)
         return write_path / read_file_relative
 
@@ -368,91 +343,66 @@ class CircuitPythonLink:
             os.remove(file_dest.resolve())
 
 
-def with_ledger(mode: str = "a"):
-    """
-    Decorator for using the ledger file; manages locking and
-    unlocking the file
-    """
-
-    def decorator_with_ledger(func):
-        """Decorator for working with the ledger file"""
-
-        @functools.wraps(func)
-        def wrapper_with_ledger(
-            entry: LedgerEntry,
-            *,
-            expect_entry: Optional[bool] = None,
-            use_lock: bool = True,
-        ) -> bool:
-            """Edit the ledger"""
-
-            # Open the ledger file
-            with open(LEDGER_FILE, mode=mode, encoding="utf-8") as filedesc:
-
-                # Use a file lock if requested
-                if use_lock:
-                    fcntl.lockf(filedesc, fcntl.LOCK_EX)
-
-                # Handle the entry
-                if (expect_entry is None) or (
-                    expect_entry == (entry.filename in iter_ledger_filenames())
-                ):
-                    result = func(entry, filedesc=filedesc)
-                else:
-                    result = False
-
-                # Release the file lock if needed
-                if use_lock:
-                    fcntl.lockf(filedesc, fcntl.LOCK_UN)
-
-                return result
-
-        return wrapper_with_ledger
-
-    return decorator_with_ledger
+def get_links_header() -> Tuple[str, ...]:
+    """Get the header row for the links list."""
+    return (
+        "ID",
+        "Name",
+        "Running?",
+        "Read Path",
+        "Write Path",
+        "Recursive?",
+        "Process ID",
+        "Base Directory",
+    )
 
 
-@with_ledger(mode="a")
-def append_to_ledger(entry: LedgerEntry, **args) -> Literal[True]:
-    """Add a file to the ledger; returns whether the file actually
-    was added (True) or if it already existed (False)
-    """
+def get_links_list(
+    pattern: str, *, abs_paths: bool = False, name: str = ""
+) -> List[_TableRowEntry]:
+    """Get the information about links."""
+    # Get the paths of all the exiting links
+    link_paths = pathlib.Path(LINKS_DIRECTORY).glob(pattern)
 
-    csvwriter = csv.writer(args["filedesc"])
-    csvwriter.writerow(entry)
-    return True
+    # Populate the list of links
+    link_infos = []
+    for link_path in link_paths:
 
+        # Load link and start getting info
+        link = CircuitPythonLink.load_link_by_filepath(str(link_path))
+        link_id = link.link_id
+        link_name = link.name
+        link_running = not link.stopped
 
-@with_ledger(mode="w")
-def remove_from_ledger(entry: LedgerEntry, **args) -> Literal[True]:
-    """Remove a file from the ledger; returns whether the file actually
-    was removed (True) or if it didn't exist (False)
-    """
+        # Attempt to use relative paths if possible
+        if not abs_paths:
+            try:
+                link_read = link.read_path.resolve().relative_to(os.getcwd())
+            except ValueError:
+                abs_paths = True
+        if abs_paths:
+            link_read = link.read_path.resolve()
 
-    csvwriter = csv.writer(args["filedesc"])
-    for existing_entry in iter_ledger_filenames(False):
-        if existing_entry != entry:
-            csvwriter.writerow(existing_entry)
-    return True
+        # Get remaining link info
+        link_write = link.write_path.resolve()
+        link_recursive = link.recursive
+        link_proc = link.process_id
+        link_base = link.base_dir
 
-
-def iter_ledger_entries(use_lock: bool = True) -> Iterator[LedgerEntry]:
-    """Iterate through ledger entries"""
-
-    with open(LEDGER_FILE, mode="r+", encoding="utf-8") as csvfile:
-        if use_lock:
-            fcntl.lockf(csvfile, fcntl.LOCK_EX)
-        csvreader = csv.reader(csvfile)
-        for ledger_entry in csvreader:
-            yield LedgerEntry(
-                ledger_entry[0], int(ledger_entry[1]), int(ledger_entry[2])
+        # If not specific named link is requested, append it
+        if not name or link_name == name:
+            link_infos.append(
+                (
+                    link_id,
+                    link_name,
+                    link_running,
+                    link_read,
+                    link_write,
+                    link_recursive,
+                    link_proc,
+                    link_base,
+                )
             )
-        if use_lock:
-            fcntl.lockf(csvfile, fcntl.LOCK_UN)
 
-
-def iter_ledger_filenames(use_lock: bool = True) -> Iterator[str]:
-    """Iterate through ledger entry filenames"""
-
-    for entry in iter_ledger_entries(use_lock):
-        yield entry.filename
+    # Return a sorted list of links
+    return sorted(link_infos, key=lambda x: x[0])
